
from __future__ import print_function

import sys
import os

import _setup_test_env  # noqa
from nose.tools import assert_equal, assert_is_instance, assert_in, assert_not_in, assert_true, assert_false
import returnn.sprint.interface as SprintAPI
from tempfile import mkdtemp
from returnn.theano.engine import Engine
from returnn.config import Config
from returnn.theano.network import LayerNetwork
import returnn.theano.util as theano_util
import shutil
import numpy


theano_util.monkey_patches()


def install_sigint_handler():
  import signal

  def signal_handler(signal, frame):
    print("\nSIGINT at:")
    better_exchook.print_tb(tb=frame, file=sys.stdout)
    print("")

    # It's likely that SIGINT was caused by Util.interrupt_main().
    # We might have a stacktrace from there.
    if getattr(sys, "exited_frame", None) is not None:
      print("interrupt_main via:")
      better_exchook.print_tb(tb=sys.exited_frame, file=sys.stdout)
      print("")
      sys.exited_frame = None
      # Normal exception instead so that Nose will catch it.
      raise Exception("Got SIGINT!")
    else:
      print("\nno sys.exited_frame\n")
      # Normal SIGINT. Normal Nose exit.
      if old_action:
        old_action()
      else:
        raise KeyboardInterrupt

  old_action = signal.signal(signal.SIGINT, signal_handler)


install_sigint_handler()


def create_first_epoch(config_filename):
  config = Config()
  config.load_file(config_filename)
  engine = Engine([])
  engine.init_train_from_config(config=config, train_data=None)
  engine.epoch = 1
  engine.save_model(engine.get_epoch_model_filename(), epoch=engine.epoch)
  Engine._epoch_model = None


def test_forward():
  tmpdir = mkdtemp("returnn-test-sprint")
  olddir = os.getcwd()
  os.chdir(tmpdir)

  open("config", "w").write(
    """
    num_inputs 2
    num_outputs 3
    hidden_size 1
    hidden_type forward
    activation relu
    bidirectional false
    model model
    log_verbosity 5
    """)

  create_first_epoch("config")

  inputDim = 2
  outputDim = 3
  SprintAPI.init(inputDim=inputDim, outputDim=outputDim,
                 config="action:forward,configfile:config,epoch:1",
                 targetMode="forward-only")
  assert isinstance(SprintAPI.engine, Engine)
  assert isinstance(SprintAPI.engine.network, LayerNetwork)
  print("used data keys via net:", SprintAPI.engine.network.get_used_data_keys())
  print("used data keys via dev:", SprintAPI.engine.devices[0].used_data_keys)

  features = numpy.array([[0.1, 0.2], [0.2, 0.3], [0.3, 0.4], [0.4, 0.5]])
  seq_len = features.shape[0]
  posteriors = SprintAPI._forward("segment1", features.T).T
  assert_equal(posteriors.shape, (seq_len, outputDim))

  SprintAPI.exit()

  os.chdir(olddir)
  shutil.rmtree(tmpdir)




"""
Collection of generic utilities related to signal processing
"""

import numpy


def greenwood_function(x, scaling_constant=165.4, constant_of_integration=0.88, slope=2.1):
  """
  Greenwood function, convert fractional length to frequency, see
  Schlueter, Ralf, et al. "Gammatone features and feature combination for large vocabulary speech recognition."
  ICASSP 2007 and also
  https://en.wikipedia.org/wiki/Greenwood_function
  The default values are taken from RASR, see https://github.com/rwth-i6/rasr/blob/master/src/Signal/GammaTone.cc
  They correspond to the recommended values for human data according to
  Greenwood, Donald D. "A cochlear frequency‐position function for several species—29 years later."
  The Journal of the Acoustical Society of America, 1990

  :param float x: fractional length
  :param float scaling_constant: A in [1]
  :param float constant_of_integration: k in [1]
  :param float slope: a in [1]
  :return: frequency corresponding to given fractional length
  :rtype: float
  """
  assert 0 <= x <= 1, "fractional length has to be between 0 and 1"
  return scaling_constant * (numpy.power(10.0, slope * x) - constant_of_integration)


def inv_greenwood_function(freq, scaling_constant=165.4, constant_of_integration=0.88, slope=2.1):
  """
  Inverse greenwood function, convert frequency to fractional length, see
  Schlueter, Ralf, et al. "Gammatone features and feature combination for large vocabulary speech recognition."
  ICASSP 2007 and also
  https://en.wikipedia.org/wiki/Greenwood_function
  The default values are taken from RASR, see https://github.com/rwth-i6/rasr/blob/master/src/Signal/GammaTone.cc
  They correspond to the recommended values for human data according to
  Greenwood, Donald D. "A cochlear frequency‐position function for several species—29 years later."
  The Journal of the Acoustical Society of America, 1990

  :param float freq: frequency
  :param float scaling_constant: A in [1]
  :param float constant_of_integration: k in [1]
  :param float slope: a in [1]
  :return: fractional length corresponding to given frequency
  :rtype: float
  """
  return numpy.log10(freq / scaling_constant + constant_of_integration) / slope


class GammatoneFilterbank(object):
  """
  Class representing a gammatone filterbank.
  Based on
  [1] Schlueter, Ralf, et al. "Gammatone features and feature combination for large vocabulary speech recognition."
  ICASSP 2007
  """

  def __init__(self, num_channels, length, sample_rate=16000, freq_max=7500., freq_min=100., normalization=True):
    """
    :param int num_channels: number of filters
    :param int length: length of FIR filters in seconds
    :param int sample_rate: sample rate of audio signal in Hz
    :param float freq_max: maximum frequency of filterbank
    :param float freq_min: minimum frequency of filterbank
    :param bool normalization: normalize filterbanks to maximum frequency response of 0 dB
    """
    self.num_channels = num_channels
    self.length = length
    self.sample_rate = sample_rate
    self.freq_max = freq_max
    self.freq_min = freq_min
    self.normalization = normalization

  def get_gammatone_filterbank(self):
    """
    Returns an array with the parameters of the gammatone filterbank

    :return: gammatone filterbank of shape (self.length * self.sample_rate, self.num_channels)
    :rtype: numpy.array
    """
    center_freqs = self.center_frequencies(self.num_channels, self.freq_max, self.freq_min)
    fbank = []
    for freq in center_freqs:
      fbank.append(self.gammatone_impulse_response(freq, self.length, self.sample_rate))
    fbank = numpy.vstack(fbank)
    if self.normalization:
      fbank = self.normalize_filters(fbank)
    fbank = numpy.transpose(fbank)
    return fbank

  @staticmethod
  def center_frequencies(num_channels, freq_max, freq_min):
    """
    Determine center frequencies for gammatone filterbank

    :param int num_channels: number of filters
    :param float freq_max: maximum frequency of filterbank
    :param float freq_min: minimum frequency of filterbank
    :return: center frequencies
    :rtype: numpy.array
    """
    x_min = inv_greenwood_function(freq_min)
    x_max = inv_greenwood_function(freq_max)
    x_linear = numpy.linspace(x_min, x_max, num_channels)
    freq_center = numpy.vectorize(greenwood_function)(x_linear)
    return freq_center

  @staticmethod
  def bandwidth_by_center_frequency(freq, lin_approx_coeff=24.7, quality_factor=9.264491981582191):
    """
    Get bandwidth (named B in [1]) by center frequency using a linear approximation of the equivalent rectangular
    bandwidth (ERB) from
    Glasberg, Brian R., and Brian CJ Moore. "Derivation of auditory filter shapes from notched-noise data."
    Hearing research, 1990
    The default values are taken from there and are also used in RASR, see
    https://github.com/rwth-i6/rasr/blob/master/src/Signal/GammaTone.cc

    :param float freq: center frequency
    :param float lin_approx_coeff: coefficient for the linear approximation of the ERB
    :param float quality_factor: audiological (ERB) based filter quality factor
    :return: bandwidth
    :rtype: float
    """
    return lin_approx_coeff * (1 / (lin_approx_coeff * quality_factor) * freq + 1.0)

  def gammatone_impulse_response(self, f_center, length, sample_rate, output_gain=1., filter_order=4, phase_shift=0.):
    """
    Compute gammatone impulse response based on [1]

    :param float f_center: center frequency
    :param int length: length of finite impulse response in seconds
    :param int sample_rate: sample rate of audio signal in Hz
    :param float output_gain: output gain, named k in [1]
    :param int filter_order: order of filter, named n in [1]
    :param float phase_shift: phase shift, named phi in [1]
    :return: gammatone impulse response
    :rtype: numpy.array
    """
    bandwidth = self.bandwidth_by_center_frequency(f_center)  # bandwidth (duration of impulse response)
    num_samples = int(numpy.floor(sample_rate * length))
    t = numpy.linspace(1. / sample_rate, length, num_samples)
    return (
      output_gain *
      numpy.power(t, filter_order - 1) *
      numpy.exp(-2 * numpy.pi * bandwidth * t) *
      numpy.cos(2 * numpy.pi * f_center * t + phase_shift))

  @staticmethod
  def normalize_filters(filters):
    """
    Normalize filterbank such that the maximum frequency response is 0 dB

    :param numpy.array filters: filterbank with shape number_channels x filter_length
    :return: normalized filterbank
    :rtype: numpy.array
    """
    from scipy import signal
    for filt in range(filters.shape[0]):
      _, f_resp = signal.freqz(filters[filt, :])
      filters[filt, :] = filters[filt, :] / numpy.max(numpy.abs(f_resp))
    return filters




from __future__ import print_function

import sys
import os

import _setup_test_env  # noqa
import sys
import numpy
import theano
import theano.tensor as T
import theano.scan_module.scan_op
import unittest
from nose.tools import assert_equal, assert_is, assert_is_instance
from numpy.testing.utils import assert_almost_equal, assert_allclose
import theano.printing
from pprint import pprint
from returnn.datasets.generating import Task12AXDataset
from returnn.theano.updater import Updater
from returnn.util.basic import have_gpu
from returnn.util.basic import NumbersDict
from returnn.config import Config
from returnn.theano.layers.hidden import DumpLayer
import returnn.__main__ as rnn
import returnn.theano.engine_util as engine_util
import returnn.theano.util as theano_util
import returnn.theano.network as network
from returnn.util import better_exchook

theano_util.monkey_patches()

# Some code uses get_global_config().
# Not sure about the most clean solution.
rnn.config = Config()


class DummyDevice:
  """
  Behave like Device.
  Only needed for assign_dev_data.
  """
  blocking = True
  used_data_keys = ("data", "classes")
  targets = None
  output_index = None

  def __init__(self):
    self.num_frames = NumbersDict(0)
    self.y = {}
    self.j = {}

  def alloc_data(self, shapes, max_ctc_length=0):
    """
    :param dict[str,list[int]] shapes: by data-key. format usually (time,batch,features)
    :type max_ctc_length: int
    """
    assert all([s > 0 for s in shapes["data"]])
    # For output_shape, we allow zeros, because e.g. in forwarding, we don't know them and will not use it.
    self.targets = {k: numpy.full(shapes[k], -1, dtype=theano.config.floatX) for k in self.used_data_keys}
    self.ctc_targets = numpy.zeros((shapes.get('classes', [0, 0])[1], max_ctc_length), dtype=theano.config.floatX)
    self.output_index = {k: numpy.zeros(shapes[k][0:2], dtype='int8') for k in self.used_data_keys}
    self.tags = [None] * shapes["data"][1]  # seq-name for each batch slice

  def initialize(self, net):
    self.y = {k: theano.shared(numpy.zeros((1,) * net.y[k].ndim, dtype=net.y[k].dtype),
                               borrow=True, name='y_%s' % k)
              for k in self.used_data_keys}
    self.j = {k: theano.shared(numpy.zeros((1, 1), dtype='int8'), borrow=True, name='j_%s' % k)
              for k in self.used_data_keys}

  def update_data(self):
    for target in self.used_data_keys:
      self.y[target].set_value(self.targets[target].astype(self.y[target].dtype), borrow = True)
    for k in self.used_data_keys:
      self.j[k].set_value(self.output_index[k], borrow = True)


def test_DummyDevice():
  dataset = Task12AXDataset(num_seqs=1000, seq_ordering="random", chunking="200:200")
  dataset.init_seq_order(epoch=1)
  batch_gen = dataset.generate_batches(recurrent_net=True, batch_size=1000, max_seqs=3)
  batches = batch_gen.peek_next_n(1)
  dev = DummyDevice()
  assign_success, _ = engine_util.assign_dev_data(device=dev, dataset=dataset, batches=batches)
  assert assign_success


def load(lstm_opts=None):
  if not lstm_opts: lstm_opts = {"class": "lstm2"}
  lstm_opts = lstm_opts.copy()
  lstm_opts.update({"n_out": 10, "from": "in"})
  num_inputs = 9
  num_outputs = 2
  net_topo = {
    "in": {"class": "dump", "filename": "in"},
    "lstm": lstm_opts,
    "lstm_dump": {"class": "dump", "from": "lstm", "filename": "lstm"},
    "output": {"class": "softmax", "loss": "ce", "from": "lstm_dump"}
  }

  collected_data = {}
  DumpLayer.global_debug_container = collected_data

  net = network.LayerNetwork.from_json(
    json_content=net_topo,
    n_in=num_inputs,
    n_out={"classes": (num_outputs, 1)},
    train_flag=True
  )
  net.declare_train_params()

  # Init dataset and prepare one minibatch.
  epoch = 1
  dataset = Task12AXDataset(num_seqs=1000, seq_ordering="random", chunking="200:200")
  dataset.init_seq_order(epoch=epoch)
  batch_gen = dataset.generate_batches(
    recurrent_net=net.recurrent,
    batch_size=5000,
    max_seqs=10)
  batches = batch_gen.peek_next_n(1)
  # We need the DummyDevice for assign_dev_data.
  dev = DummyDevice()
  assign_success, _ = engine_util.assign_dev_data(device=dev, dataset=dataset, batches=batches)
  assert assign_success
  dev.initialize(net)
  dev.update_data()
  givens = [(net.y[k], dev.y[k]) for k in dev.used_data_keys]
  givens += [(net.j[k], dev.j[k]) for k in dev.used_data_keys]

  # Now gradients, updates and compile everything.
  gradients = {p: T.grad(net.get_objective(), p, known_grads=net.known_grads)
               for p in net.train_params_vars}
  updater = Updater(adam=True)
  updater.initVars(net, gradients)
  updater.setLearningRate(learning_rate=0.01)
  trainer = theano.function(
    inputs=[],
    outputs=[net.total_cost],
    givens=givens,
    updates=updater.getUpdateList(),
    on_unused_input='warn',
    name="train_and_updater")

  for p in net.train_params_vars:
    collected_data["param:%s" % p.name] = p.get_value()

  # And finally, run it.
  cost = trainer()
  collected_data["cost"] = cost
  return collected_data


def test_load():
  load()


atol = 1e-7


def compare_lstm(lstm_opts=None):
  res1 = load()
  res2 = load(lstm_opts=lstm_opts)
  fail = False
  print("keys in res1:", sorted(res1.keys()))
  print("keys in res2:", sorted(res2.keys()))
  for key in sorted(res1.keys()):
    if key not in res2:
      print("ERROR: %r not in res2" % key)
      fail = True
    v1 = res1[key]
    v2 = res2[key]
    v1 = numpy.asarray(v1)
    v2 = numpy.asarray(v2)
    if v1.shape != v2.shape:
      print("shape does not match for %r" % key)
      print("v1 shape:", v1.shape)
      print("v2 shape:", v2.shape)
      fail = True
    elif not numpy.allclose(v1, v2, atol=atol):
      print("not equal: %r" % key)
      c = 0
      for idx in zip(*numpy.unravel_index(range(numpy.prod(v1.shape)), v1.shape)):
        e1 = v1[idx]
        e2 = v2[idx]
        if not numpy.isclose(e1, e2, atol=atol):
          print("idx %r differs: %r vs %r" % (idx, e1, e2))
          c += 1
          if c >= 10: break
      fail = True
  for key in sorted(res2.keys()):
    if key not in res1:
      print("ERROR: %r not in res1" % key)
      fail = True
  assert not fail


def test_native_lstm():
  compare_lstm({"class": "native_lstm"})


@unittest.skipIf(not have_gpu(), "no gpu on this system")
def test_fast_bw():
  print("Make op...")
  from returnn.native_op import FastBaumWelchOp
  op = FastBaumWelchOp().make_theano_op()  # (am_scores, edges, weights, start_end_states, float_idx, state_buffer)
  print("Op:", op)
  n_batch = 3
  seq_len = 5
  n_classes = 5
  from returnn.util.fsa import FastBwFsaShared
  fsa = FastBwFsaShared()
  fsa.add_inf_loop(state_idx=0, num_emission_labels=n_classes)
  fast_bw_fsa = fsa.get_fast_bw_fsa(n_batch=n_batch)
  edges = fast_bw_fsa.edges.view("float32")
  edges_placeholder = T.fmatrix(name="edges")
  weights = fast_bw_fsa.weights
  weights_placeholder = T.fvector(name="weights")
  start_end_states = fast_bw_fsa.start_end_states.view("float32")
  start_end_states_placeholder = T.fmatrix(name="start_end_states")
  am_scores = numpy.random.normal(size=(seq_len, n_batch, n_classes)).astype("float32")  # in -log space
  am_scores_placeholder = T.ftensor3(name="am_scores")
  float_idx = numpy.ones((seq_len, n_batch), dtype="float32")
  float_idx_placeholder = T.fmatrix(name="float_idx")
  last_state_idx = numpy.max(fast_bw_fsa.start_end_states[1])  # see get_automata_for_batch
  state_buffer = numpy.zeros((2, last_state_idx + 1), dtype="float32")
  state_buffer_placeholder = T.fmatrix(name="state_buffer")
  print("Construct call...")
  fwdbwd, obs_scores = op(
    am_scores_placeholder, edges_placeholder, weights_placeholder, start_end_states_placeholder, float_idx_placeholder, state_buffer_placeholder)
  f = theano.function(inputs=[am_scores_placeholder, edges_placeholder, weights_placeholder, start_end_states_placeholder, float_idx_placeholder, state_buffer_placeholder], outputs=[fwdbwd, obs_scores])
  print("Done.")
  print("Eval:")
  _, score = f(am_scores, edges, weights, start_end_states, float_idx, state_buffer)
  print("score:", score)


@unittest.skipIf(not have_gpu(), "no gpu on this system")
def test_fast_bw_uniform():
  print("Make op...")
  from returnn.native_op import FastBaumWelchOp
  op = FastBaumWelchOp().make_theano_op()  # (am_scores, edges, weights, start_end_states, float_idx, state_buffer)
  print("Op:", op)
  n_batch = 3
  seq_len = 7
  n_classes = 5
  from returnn.util.fsa import FastBwFsaShared
  fsa = FastBwFsaShared()
  for i in range(n_classes):
    fsa.add_edge(i, i + 1, emission_idx=i)  # fwd
    fsa.add_edge(i + 1, i + 1, emission_idx=i)  # loop
  assert n_classes <= seq_len
  fast_bw_fsa = fsa.get_fast_bw_fsa(n_batch=n_batch)
  print("edges:")
  print(fast_bw_fsa.edges)
  edges = fast_bw_fsa.edges.view("float32")
  edges_placeholder = T.fmatrix(name="edges")
  weights = fast_bw_fsa.weights
  weights_placeholder = T.fvector(name="weights")
  print("start_end_states:")
  print(fast_bw_fsa.start_end_states)
  start_end_states = fast_bw_fsa.start_end_states.view("float32")
  start_end_states_placeholder = T.fmatrix(name="start_end_states")
  am_scores = numpy.ones((seq_len, n_batch, n_classes), dtype="float32") * numpy.float32(1.0 / n_classes)
  am_scores = -numpy.log(am_scores)  # in -log space
  am_scores_placeholder = T.ftensor3(name="am_scores")
  float_idx = numpy.ones((seq_len, n_batch), dtype="float32")
  float_idx_placeholder = T.fmatrix(name="float_idx")
  last_state_idx = numpy.max(fast_bw_fsa.start_end_states[1])  # see get_automata_for_batch
  state_buffer = numpy.zeros((2, last_state_idx + 1), dtype="float32")
  state_buffer_placeholder = T.fmatrix(name="state_buffer")
  print("Construct call...")
  fwdbwd, obs_scores = op(
    am_scores_placeholder, edges_placeholder, weights_placeholder, start_end_states_placeholder, float_idx_placeholder, state_buffer_placeholder)
  f = theano.function(inputs=[am_scores_placeholder, edges_placeholder, weights_placeholder, start_end_states_placeholder, float_idx_placeholder, state_buffer_placeholder], outputs=[fwdbwd, obs_scores])
  print("Done.")
  print("Eval:")
  fwdbwd, score = f(am_scores, edges, weights, start_end_states, float_idx, state_buffer)
  print("score:")
  print(repr(score))
  assert_equal(score.shape, (seq_len, n_batch))
  bw = numpy.exp(-fwdbwd)
  print("Baum-Welch soft alignment:")
  print(repr(bw))
  assert_equal(bw.shape, (seq_len, n_batch, n_classes))
  from numpy import array, float32
  if seq_len == n_classes:
    print("Extra check identity...")
    for i in range(n_batch):
      assert_almost_equal(numpy.identity(n_classes), bw[:, i])
  if seq_len == 7 and n_classes == 5:
    print("Extra check ref_align (7,5)...")
    assert_allclose(score, 8.55801582, rtol=1e-5)  # should be the same everywhere
    ref_align = \
      array([[[1., 0., 0., 0., 0.]],
             [[0.33333316, 0.66666663, 0., 0., 0.]],
             [[0.06666669, 0.53333354, 0.40000018, 0., 0.]],
             [[0., 0.20000014, 0.60000014, 0.19999999, 0.]],
             [[0., 0., 0.39999962, 0.53333312, 0.06666663]],
             [[0., 0., 0., 0.66666633, 0.33333316]],
             [[0., 0., 0., 0., 0.99999982]]], dtype=float32)
    assert_equal(ref_align.shape, (seq_len, 1, n_classes))
    ref_align = numpy.tile(ref_align, (1, n_batch, 1))
    assert_equal(ref_align.shape, bw.shape)
    # print("Reference alignment:")
    # print(repr(ref_align))
    print("mean square diff:", numpy.mean(numpy.square(ref_align - bw)))
    print("max square diff:", numpy.max(numpy.square(ref_align - bw)))
    assert_allclose(ref_align, bw, rtol=1e-5)
  print("Done.")


if __name__ == "__main__":
  try:
    better_exchook.install()
    if len(sys.argv) <= 1:
      for k, v in sorted(globals().items()):
        if k.startswith("test_"):
          print("-" * 40)
          print("Executing: %s" % k)
          try:
            v()
          except unittest.SkipTest as exc:
            print("SkipTest:", exc)
          print("-" * 40)
      print("Finished all tests.")
    else:
      assert len(sys.argv) >= 2
      for arg in sys.argv[1:]:
        print("Executing: %s" % arg)
        if arg in globals():
          globals()[arg]()  # assume function and execute
        else:
          eval(arg)  # assume Python code and execute
  finally:
    import threading
    #if len(list(threading.enumerate())) > 1:
    #  print("Warning, more than one thread at exit:")
    #  better_exchook.dump_all_thread_tracebacks()





# Also see test_SprintDataset.py.

from __future__ import print_function

import sys
import _setup_test_env  # noqa
import unittest
import numpy
from nose.tools import assert_equal, assert_is_instance, assert_in, assert_not_in, assert_true, assert_false
from returnn.datasets.generating import GeneratingDataset, DummyDataset, DummyDatasetMultipleSequenceLength
from returnn.engine.batch import Batch
from returnn.datasets.basic import Dataset
from returnn.util.basic import NumbersDict

from returnn.util import better_exchook


def test_generate_batches():
  dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=20)
  dataset.init_seq_order(1)
  batch_gen = dataset.generate_batches(recurrent_net=False, max_seqs=2, batch_size=5)
  while batch_gen.has_more():
    batch_gen.peek_next_n(1)
    batch_gen.advance(1)


def test_generate_batches_recurrent():
  dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=20)
  dataset.init_seq_order(1)
  batch_gen = dataset.generate_batches(recurrent_net=True, max_seqs=2, batch_size=5)
  while batch_gen.has_more():
    batch_gen.peek_next_n(1)
    batch_gen.advance(1)


def test_iterate_seqs_no_chunking_1():
  dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=2, seq_len=11)
  dataset.init_seq_order(1)
  seqs = list(dataset.iterate_seqs(chunk_size=0, chunk_step=0, used_data_keys=None))
  assert_equal(len(seqs), 2)
  assert_equal(seqs[0], (0, 0, 11))  # seq-idx, start-frame, end-frame
  assert_equal(seqs[1], (1, 0, 11))


def test_iterate_seqs_chunking_1():
  dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=2, seq_len=11)
  dataset.init_seq_order(1)
  seqs = list(dataset.iterate_seqs(chunk_size=10, chunk_step=5, used_data_keys=None))
  for s in seqs:
    print(s)
  assert_equal(len(seqs), 6)
  assert_equal(seqs[0], (0, 0, 10))  # seq-idx, start-frame, end-frame
  assert_equal(seqs[1], (0, 5, 11))
  assert_equal(seqs[2], (0, 10, 11))
  assert_equal(seqs[3], (1, 0, 10))
  assert_equal(seqs[4], (1, 5, 11))
  assert_equal(seqs[5], (1, 10, 11))


def test_iterate_seqs_chunking_varying_sequence_length():
  dataset = DummyDatasetMultipleSequenceLength(input_dim=2, output_dim=3, num_seqs=2, seq_len={'data': 24, 'classes': 12})
  dataset.init_seq_order(1)
  seqs = list(dataset.iterate_seqs(chunk_size={'data': 12, 'classes': 6}, chunk_step={'data': 6, 'classes': 3}, used_data_keys=None))
  for s in seqs:
    print(s)
  assert_equal(len(seqs), 8)
  assert_equal(seqs[0], (0, NumbersDict({'data':0, 'classes': 0}), NumbersDict({'data':12, 'classes': 6})))
  assert_equal(seqs[1], (0, NumbersDict({'data':6, 'classes': 3}), NumbersDict({'data':18, 'classes': 9})))
  assert_equal(seqs[2], (0, NumbersDict({'data':12, 'classes': 6}), NumbersDict({'data':24, 'classes': 12})))
  assert_equal(seqs[3], (0, NumbersDict({'data':18, 'classes': 9}), NumbersDict({'data':24, 'classes': 12})))
  assert_equal(seqs[4], (1, NumbersDict({'data':0, 'classes': 0}), NumbersDict({'data':12, 'classes': 6})))
  assert_equal(seqs[5], (1, NumbersDict({'data':6, 'classes': 3}), NumbersDict({'data':18, 'classes': 9})))
  assert_equal(seqs[6], (1, NumbersDict({'data':12, 'classes': 6}), NumbersDict({'data':24, 'classes': 12})))
  assert_equal(seqs[7], (1, NumbersDict({'data':18, 'classes': 9}), NumbersDict({'data':24, 'classes': 12})))


def test_batches_recurrent_1():
  dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=2, seq_len=11)
  dataset.init_seq_order(1)
  dataset.chunk_size = 10
  dataset.chunk_step = 5
  batch_gen = dataset.generate_batches(recurrent_net=True, max_seqs=1, batch_size=20)
  all_batches = []; " :type: list[Batch] "
  while batch_gen.has_more():
    batch, = batch_gen.peek_next_n(1)
    assert_is_instance(batch, Batch)
    print("batch:", batch)
    print("batch seqs:", batch.seqs)
    all_batches.append(batch)
    batch_gen.advance(1)

  # Each batch will have 1 batch-slice (max_seqs) and up to 10 frames (chunk_size).
  # For each seq, we get 3 chunks (chunk_step 5 for 11 frames).
  # Thus, 6 batches.
  assert_equal(len(all_batches), 6)

  assert_equal(all_batches[0].start_seq, 0)
  assert_equal(all_batches[0].end_seq, 1)  # exclusive
  assert_equal(len(all_batches[0].seqs), 1)  # 1 BatchSeqCopyPart
  assert_equal(all_batches[0].seqs[0].seq_idx, 0)
  assert_equal(all_batches[0].seqs[0].seq_start_frame, 0)
  assert_equal(all_batches[0].seqs[0].seq_end_frame, 10)
  assert_equal(all_batches[0].seqs[0].frame_length, 10)
  assert_equal(all_batches[0].seqs[0].batch_slice, 0)
  assert_equal(all_batches[0].seqs[0].batch_frame_offset, 0)

  assert_equal(all_batches[1].start_seq, 0)
  assert_equal(all_batches[1].end_seq, 1)  # exclusive
  assert_equal(len(all_batches[1].seqs), 1)  # 1 BatchSeqCopyPart
  assert_equal(all_batches[1].seqs[0].seq_idx, 0)
  assert_equal(all_batches[1].seqs[0].seq_start_frame, 5)
  assert_equal(all_batches[1].seqs[0].seq_end_frame, 11)
  assert_equal(all_batches[1].seqs[0].frame_length, 6)
  assert_equal(all_batches[1].seqs[0].batch_slice, 0)
  assert_equal(all_batches[1].seqs[0].batch_frame_offset, 0)

  assert_equal(all_batches[2].start_seq, 0)
  assert_equal(all_batches[2].end_seq, 1)  # exclusive
  assert_equal(len(all_batches[2].seqs), 1)  # 1 BatchSeqCopyPart
  assert_equal(all_batches[2].seqs[0].seq_idx, 0)
  assert_equal(all_batches[2].seqs[0].seq_start_frame, 10)
  assert_equal(all_batches[2].seqs[0].seq_end_frame, 11)
  assert_equal(all_batches[2].seqs[0].frame_length, 1)
  assert_equal(all_batches[2].seqs[0].batch_slice, 0)
  assert_equal(all_batches[2].seqs[0].batch_frame_offset, 0)

  assert_equal(all_batches[3].start_seq, 1)
  assert_equal(all_batches[3].end_seq, 2)  # exclusive
  assert_equal(len(all_batches[3].seqs), 1)  # 1 BatchSeqCopyPart
  assert_equal(all_batches[3].seqs[0].seq_idx, 1)
  assert_equal(all_batches[3].seqs[0].seq_start_frame, 0)
  assert_equal(all_batches[3].seqs[0].seq_end_frame, 10)
  assert_equal(all_batches[3].seqs[0].frame_length, 10)
  assert_equal(all_batches[3].seqs[0].batch_slice, 0)
  assert_equal(all_batches[3].seqs[0].batch_frame_offset, 0)

  # ...


def test_batches_non_recurrent_1():
  dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=2, seq_len=11)
  dataset.init_seq_order(1)
  batch_gen = dataset.generate_batches(recurrent_net=False, max_seqs=2, batch_size=5)
  all_batches = []  # type: list[Batch]
  while batch_gen.has_more():
    batch, = batch_gen.peek_next_n(1)
    assert_is_instance(batch, Batch)
    print("batch:", batch)
    print("batch seqs:", batch.seqs)
    all_batches.append(batch)
    batch_gen.advance(1)

  # Each batch will have 5 frames (batch_size), not more, i.e. a single seq.
  # There are 2 * 11 frames in total, so 5 batches, because we concat the 2 seqs, in the non-recurrent case.
  assert_equal(len(all_batches), 5)

  assert_equal(all_batches[0].start_seq, 0)
  assert_equal(all_batches[0].end_seq, 1)  # exclusive
  assert_equal(len(all_batches[0].seqs), 1)  # 1 BatchSeqCopyPart
  assert_equal(all_batches[0].seqs[0].seq_idx, 0)
  assert_equal(all_batches[0].seqs[0].seq_start_frame, 0)
  assert_equal(all_batches[0].seqs[0].seq_end_frame, 5)
  assert_equal(all_batches[0].seqs[0].frame_length, 5)
  assert_equal(all_batches[0].seqs[0].batch_slice, 0)
  assert_equal(all_batches[0].seqs[0].batch_frame_offset, 0)

  assert_equal(all_batches[1].start_seq, 0)
  assert_equal(all_batches[1].end_seq, 1)  # exclusive
  assert_equal(len(all_batches[1].seqs), 1)  # 1 BatchSeqCopyPart
  assert_equal(all_batches[1].seqs[0].seq_idx, 0)
  assert_equal(all_batches[1].seqs[0].seq_start_frame, 5)
  assert_equal(all_batches[1].seqs[0].seq_end_frame, 10)
  assert_equal(all_batches[1].seqs[0].frame_length, 5)
  assert_equal(all_batches[1].seqs[0].batch_slice, 0)
  assert_equal(all_batches[1].seqs[0].batch_frame_offset, 0)

  assert_equal(all_batches[2].start_seq, 0)
  assert_equal(all_batches[2].end_seq, 2)  # exclusive. now both seq 0 and 1
  assert_equal(len(all_batches[2].seqs), 2)  # two copies, BatchSeqCopyPart
  assert_equal(all_batches[2].seqs[0].seq_idx, 0)
  assert_equal(all_batches[2].seqs[0].seq_start_frame, 10)
  assert_equal(all_batches[2].seqs[0].seq_end_frame, 11)
  assert_equal(all_batches[2].seqs[0].frame_length, 1)
  assert_equal(all_batches[2].seqs[0].batch_slice, 0)
  assert_equal(all_batches[2].seqs[0].batch_frame_offset, 0)
  assert_equal(all_batches[2].seqs[1].seq_idx, 1)
  assert_equal(all_batches[2].seqs[1].seq_start_frame, 0)
  assert_equal(all_batches[2].seqs[1].seq_end_frame, 4)
  assert_equal(all_batches[2].seqs[1].frame_length, 4)
  assert_equal(all_batches[2].seqs[1].batch_slice, 0)
  assert_equal(all_batches[2].seqs[1].batch_frame_offset, 1)

  assert_equal(all_batches[3].start_seq, 1)
  assert_equal(all_batches[3].end_seq, 2)  # exclusive
  assert_equal(len(all_batches[3].seqs), 1)  # 1 BatchSeqCopyPart
  assert_equal(all_batches[3].seqs[0].seq_idx, 1)
  assert_equal(all_batches[3].seqs[0].seq_start_frame, 4)
  assert_equal(all_batches[3].seqs[0].seq_end_frame, 9)
  assert_equal(all_batches[3].seqs[0].frame_length, 5)
  assert_equal(all_batches[3].seqs[0].batch_slice, 0)
  assert_equal(all_batches[3].seqs[0].batch_frame_offset, 0)

  assert_equal(all_batches[4].start_seq, 1)
  assert_equal(all_batches[4].end_seq, 2)  # exclusive
  assert_equal(len(all_batches[4].seqs), 1)  # 1 BatchSeqCopyPart
  assert_equal(all_batches[4].seqs[0].seq_idx, 1)
  assert_equal(all_batches[4].seqs[0].seq_start_frame, 9)
  assert_equal(all_batches[4].seqs[0].seq_end_frame, 11)
  assert_equal(all_batches[4].seqs[0].frame_length, 2)
  assert_equal(all_batches[4].seqs[0].batch_slice, 0)
  assert_equal(all_batches[4].seqs[0].batch_frame_offset, 0)


def test_batches_context_window():
  context_window = 2
  ctx_lr = context_window - 1
  ctx_left = ctx_lr // 2
  ctx_right = ctx_lr - ctx_left

  dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=1, seq_len=11, context_window=context_window)
  dataset.init_seq_order(1)
  dataset.chunk_size = 5
  dataset.chunk_step = 5
  batch_gen = dataset.generate_batches(recurrent_net=True, max_seqs=1, batch_size=20)
  all_batches = []  # type: list[Batch]
  while batch_gen.has_more():
    batch, = batch_gen.peek_next_n(1)
    assert_is_instance(batch, Batch)
    print("batch:", batch)
    print("batch seqs:", batch.seqs)
    all_batches.append(batch)
    batch_gen.advance(1)

  # Each batch will have 1 batch-slice (max_seqs) and up to 10 frames (chunk_size).
  # For each seq, we get 3 chunks (chunk_step 5 for 11 frames).
  # Thus, 3 batches.
  assert_equal(len(all_batches), 3)
  b0, b1, b2 = all_batches
  assert isinstance(b0, Batch)
  assert isinstance(b1, Batch)
  assert isinstance(b2, Batch)

  assert_equal(b0.start_seq, 0)
  assert_equal(b0.end_seq, 1)  # exclusive
  assert_equal(len(b0.seqs), 1)  # 1 BatchSeqCopyPart
  assert_equal(b0.seqs[0].seq_idx, 0)
  assert_equal(b0.seqs[0].seq_start_frame["classes"], 0)
  assert_equal(b0.seqs[0].seq_end_frame["classes"], 5)
  assert_equal(b0.seqs[0].frame_length["classes"], 5)
  assert_equal(b0.seqs[0].seq_start_frame["data"], 0 - ctx_left)
  assert_equal(b0.seqs[0].seq_end_frame["data"], 5 + ctx_right)
  assert_equal(b0.seqs[0].frame_length["data"], 5 + ctx_lr)
  assert_equal(b0.seqs[0].batch_slice, 0)
  assert_equal(b0.seqs[0].batch_frame_offset, 0)

  assert_equal(b1.start_seq, 0)
  assert_equal(b1.end_seq, 1)  # exclusive
  assert_equal(len(b1.seqs), 1)  # 1 BatchSeqCopyPart
  assert_equal(b1.seqs[0].seq_idx, 0)
  assert_equal(b1.seqs[0].seq_start_frame["classes"], 5)
  assert_equal(b1.seqs[0].seq_end_frame["classes"], 10)
  assert_equal(b1.seqs[0].frame_length["classes"], 5)
  assert_equal(b1.seqs[0].seq_start_frame["data"], 5 - ctx_left)
  assert_equal(b1.seqs[0].seq_end_frame["data"], 10 + ctx_right)
  assert_equal(b1.seqs[0].frame_length["data"], 5 + ctx_lr)
  assert_equal(b1.seqs[0].batch_slice, 0)
  assert_equal(b1.seqs[0].batch_frame_offset, 0)

  assert_equal(b2.start_seq, 0)
  assert_equal(b2.end_seq, 1)  # exclusive
  assert_equal(len(b2.seqs), 1)  # 1 BatchSeqCopyPart
  assert_equal(b2.seqs[0].seq_idx, 0)
  assert_equal(b2.seqs[0].seq_start_frame["classes"], 10)
  assert_equal(b2.seqs[0].seq_end_frame["classes"], 11)
  assert_equal(b2.seqs[0].frame_length["classes"], 1)
  assert_equal(b2.seqs[0].seq_start_frame["data"], 10 - ctx_left)
  assert_equal(b2.seqs[0].seq_end_frame["data"], 11 + ctx_right)
  assert_equal(b2.seqs[0].frame_length["data"], 1 + ctx_lr)
  assert_equal(b2.seqs[0].batch_slice, 0)
  assert_equal(b2.seqs[0].batch_frame_offset, 0)


def test_task12ax_window():
  from returnn.datasets.generating import Task12AXDataset
  window = 3
  dataset_kwargs = dict(num_seqs=10)
  dataset1 = Task12AXDataset(**dataset_kwargs)
  dataset2 = Task12AXDataset(window=window, **dataset_kwargs)
  input_dim = dataset1.num_inputs
  dataset1.initialize()
  dataset2.initialize()
  dataset1.init_seq_order(epoch=1)
  dataset2.init_seq_order(epoch=1)
  dataset1.load_seqs(0, 1)
  dataset2.load_seqs(0, 1)
  assert_equal(dataset1.get_data_dim("data"), input_dim)
  assert_equal(dataset2.get_data_dim("data"), input_dim * window)
  data1 = dataset1.get_data(0, "data")
  data2 = dataset2.get_data(0, "data")
  seq_len = data1.shape[0]
  assert_equal(data1.shape, (seq_len, input_dim))
  assert_equal(data2.shape, (seq_len, window * input_dim))
  data2a = data2.reshape(seq_len, window, input_dim)
  print("data1:")
  print(data1)
  print("data2:")
  print(data2)
  print("data1[0]:")
  print(data1[0])
  print("data2[0]:")
  print(data2[0])
  print("data2a[0,0]:")
  print(data2a[0, 0])
  assert_equal(list(data2a[0, 0]), [0] * input_dim)  # zero-padded left
  assert_equal(list(data2a[0, 1]), list(data1[0]))
  assert_equal(list(data2a[0, 2]), list(data1[1]))
  assert_equal(list(data2a[1, 0]), list(data1[0]))
  assert_equal(list(data2a[1, 1]), list(data1[1]))
  assert_equal(list(data2a[1, 2]), list(data1[2]))
  assert_equal(list(data2a[-1, 2]), [0] * input_dim)  # zero-padded right


def test_get_seq_order():
  dataset = Dataset()
  num_seqs = 30

  def get_seq_len(i):
    return i ** 2 % 17  # some dummy lengths

  for seq_ordering in [
      "default", "default_every_n:5", "sorted", "sorted_reverse", "random:3", "laplace:3", "laplace:.10",
      "sort_bin_shuffle:3", "sort_bin_shuffle_x2:.10"]:

    dataset.seq_ordering = seq_ordering

    # test full epoch
    dataset.partition_epoch = 1
    epoch = 3
    seq_index = dataset.get_seq_order_for_epoch(epoch, num_seqs, get_seq_len)

    assert isinstance(seq_index, (list, range, numpy.ndarray))
    assert len(set(seq_index)) == num_seqs  # right number of sequences, no duplicates

    # test partitioned epoch
    partition_epoch = 4
    dataset.partition_epoch = partition_epoch
    all_partitions_seq_index = []
    for epoch in range(1, partition_epoch + 1):
      partition_seq_index = dataset.get_seq_order_for_epoch(epoch, num_seqs, get_seq_len)
      all_partitions_seq_index += list(partition_seq_index)

    # Make sure partitions combined result in full epoch. This tests the random seed of Dataset which should be
    # fixed across partitions.
    assert set(all_partitions_seq_index) == set(seq_index)


if __name__ == "__main__":
  better_exchook.install()
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute
